/*
 *    Common definitions
 *
 * Copyright 2020 Ariel Darshan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#ifndef __NCRYPT_INTERNAL__
#define __NCRYPT_INTERNAL__

#include "bcrypt.h"
#include "wine/list.h"

typedef struct _NCRYPT_KEY_STORAGE_FUNCTION_TABLE
{
	BCRYPT_INTERFACE_VERSION version;
	SECURITY_STATUS(*OpenProvider)(NCRYPT_PROV_HANDLE *outProvider, LPCWSTR name, DWORD flags);
	SECURITY_STATUS(*OpenKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE *outKey, LPCWSTR keyName, DWORD legacyKeySpec, DWORD flags);
	SECURITY_STATUS(*CreatePersistedKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE* outKey, LPCWSTR algId, LPCWSTR keyName, DWORD legacyKeySpec, DWORD flags);
	SECURITY_STATUS(*GetProviderProperty)(NCRYPT_PROV_HANDLE provider, LPCWSTR propertyName, BYTE *output, DWORD outputSize, DWORD *result, DWORD flags);
	SECURITY_STATUS(*GetKeyProperty)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, LPCWSTR propertyName, BYTE* output, DWORD outputSize, DWORD* result, DWORD flags);
	SECURITY_STATUS(*SetProviderProperty)(NCRYPT_PROV_HANDLE provider, LPCWSTR propertyName, BYTE* input, DWORD inputSize, DWORD flags);
	SECURITY_STATUS(*SetKeyProperty)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, LPCWSTR propertyName, BYTE* input, DWORD inputSize, DWORD flags);
	SECURITY_STATUS(*FinalizeKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, DWORD flags);
	SECURITY_STATUS(*DeleteKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, DWORD flags);
	SECURITY_STATUS(*FreeProvider)(NCRYPT_PROV_HANDLE provider);
	SECURITY_STATUS(*FreeKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key);
	SECURITY_STATUS(*FreeBuffer)(void *buffer);
	SECURITY_STATUS(*Encrypt)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, BYTE *input, DWORD inputSize, void *paddingInfo, BYTE *output, DWORD outputSize, DWORD *result, DWORD flags);
	SECURITY_STATUS(*Decrypt)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, BYTE* input, DWORD inputSize, void *paddingInfo, BYTE* output, DWORD outputSize, DWORD* result, DWORD flags);
	SECURITY_STATUS(*IsAlgSupported)(NCRYPT_PROV_HANDLE provider, LPCWSTR algId, DWORD flags);
	SECURITY_STATUS(*EnumAlgorithms)(NCRYPT_PROV_HANDLE provider, DWORD algOperations, DWORD *algCount, NCryptAlgorithmName **algList, DWORD flags);
	SECURITY_STATUS(*EnumKeys)(NCRYPT_PROV_HANDLE provider, LPCWSTR scope, NCryptKeyName **keyName, void **enumState, DWORD flags);
	SECURITY_STATUS(*ImportKey)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE decryptKey, LPCWSTR blobType, NCryptBufferDesc *parameterList, NCRYPT_KEY_HANDLE *outKey, BYTE *data, DWORD dataSize, DWORD flags);
	SECURITY_STATUS(*ExportKey)(NCRYPT_KEY_HANDLE key, NCRYPT_KEY_HANDLE encryptKey, LPCWSTR blobType, NCryptBufferDesc *parameterList, BYTE *output, DWORD outputSize, DWORD *result, DWORD flags);
	SECURITY_STATUS(*SignHash)(NCRYPT_KEY_HANDLE key, void *paddingInfo, BYTE *hashValue, DWORD hashSize, BYTE *outSig, DWORD outSigSize, DWORD *result, DWORD flags);
	SECURITY_STATUS(*VerifySignature)(NCRYPT_KEY_HANDLE key, void *paddingInfo, BYTE *hashValue, DWORD hashSize, BYTE *signature, DWORD signatureSize, DWORD flags);
	SECURITY_STATUS(*PromptUser)(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE key, LPCWSTR operation, DWORD flags);
	SECURITY_STATUS(*NotifyChangeKey)(NCRYPT_PROV_HANDLE provider, HANDLE *event, DWORD flags);
	SECURITY_STATUS(*SecretAgreement)(NCRYPT_KEY_HANDLE privateKey, NCRYPT_KEY_HANDLE publicKey, NCRYPT_SECRET_HANDLE *agreedSecret, DWORD flags);
	SECURITY_STATUS(*DeriveKey)(NCRYPT_SECRET_HANDLE sharedSecret, LPCWSTR kdf, NCryptBufferDesc *parameterList, BYTE *outKey, DWORD outKeySize, DWORD *result, DWORD flags);
	SECURITY_STATUS(*FreeSecret)(NCRYPT_PROV_HANDLE provider, NCRYPT_SECRET_HANDLE sharedSecret);

}NCRYPT_KEY_STORAGE_FUNCTION_TABLE;

typedef NTSTATUS(*getInterfaceFn)(LPCWSTR providerName, NCRYPT_KEY_STORAGE_FUNCTION_TABLE **outFunctionTable, DWORD dwFlags);


#define NCRYPT_OBJ_TYPE_PROVIDER    0x44440001
#define NCRYPT_OBJ_TYPE_KEY         0x44440002
#define NCRYPT_OBJ_TYPE_HASH        0x44440003
#define NCRYPT_OBJ_TYPE_SECRET      0x44440004

struct ncrypt_base_object
{
    DWORD sType;
};

struct ncrypt_provider_instance
{
    DWORD sType;
    DWORD refCount;
    HMODULE kspDLL;
    NCRYPT_KEY_STORAGE_FUNCTION_TABLE functions;
    NCRYPT_PROV_HANDLE kspHandle;
    WCHAR *name;
};

struct ncrypt_key_instance
{
    DWORD sType;
    NCRYPT_PROV_HANDLE provider;
    NCRYPT_KEY_HANDLE kspHandle;
    WCHAR *name;
};


#define handle2baseObject(x) ((struct ncrypt_base_object*)(x))
#define baseObject2handle(x) ((NCRYPT_HANDLE)(x))

#define handle2key(x) ((struct ncrypt_key_instance*)(x))
#define key2handle(x) ((NCRYPT_KEY_HANDLE)(x))

#define handle2provider(x) ((struct ncrypt_provider_instance*)(x))
#define provider2handle(x) ((NCRYPT_PROV_HANDLE)(x))

/* A buffer should be passed to the provider that allocated it,
	so we need to keep track of it */
struct ncrypt_buffer
{
	struct list entry;
	NCRYPT_PROV_HANDLE owner;
	const void *buffer;
};


static SECURITY_STATUS open_provider(NCRYPT_PROV_HANDLE *providerInstance ,const WCHAR *name, const WCHAR *dllName, DWORD flags);
static struct ncrypt_key_instance* create_key(NCRYPT_PROV_HANDLE provider, const WCHAR *name);
static SECURITY_STATUS register_buffer(NCRYPT_PROV_HANDLE provider, const void *buffer);
static NCRYPT_PROV_HANDLE ref_provider(NCRYPT_PROV_HANDLE provider);
static void unref_provider(NCRYPT_PROV_HANDLE provider);
static SECURITY_STATUS free_provider(NCRYPT_HANDLE provider);
static SECURITY_STATUS free_key(NCRYPT_HANDLE key);

#endif /* __NCRYPT_INTERNAL__ */
