/*
 * Copyright 2020 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#if 0
#pragma makedep unix
#endif

#include "config.h"
#include "wine/port.h"

#include <assert.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#define COBJMACROS

#include "ntstatus.h"
#define WIN32_NO_STATUS
#include "windef.h"
#include "winbase.h"
#include "winerror.h"
#include "winternl.h"
#include "objidl.h"

#ifdef SONAME_LIBJXRGLUE
#define ERR JXR_ERR
#include <JXRGlue.h>
#undef ERR
#endif

#include "wine/debug.h"

#include "wmphoto_private.h"

WINE_DEFAULT_DEBUG_CHANNEL(wincodecs);

#ifdef SONAME_LIBJXRGLUE
static void *libjxrglue;
static typeof(PKImageDecode_Create_WMP) *pPKImageDecode_Create_WMP;

struct decoder_proxy
{
    struct jxrlib_decoder decoder_iface;
    struct WMPStream WMPStream_iface;
    struct jxrlib_stream *stream;
    PKImageDecode *decoder;
};

static inline struct decoder_proxy *impl_from_jxrlib_decoder(struct jxrlib_decoder *iface)
{
    return CONTAINING_RECORD(iface, struct decoder_proxy, decoder_iface);
}

static inline struct decoder_proxy *impl_from_WMPStream(struct WMPStream *iface)
{
    return CONTAINING_RECORD(iface, struct decoder_proxy, WMPStream_iface);
}

static JXR_ERR wmp_stream_Close(struct WMPStream **piface)
{
    TRACE("iface %p.\n", piface);
    return WMP_errSuccess;
}

static Bool wmp_stream_EOS(struct WMPStream *iface)
{
    FIXME("iface %p, stub!\n", iface);
    return FALSE;
}

static JXR_ERR wmp_stream_Read(struct WMPStream *iface, void *buf, size_t len)
{
    struct decoder_proxy *This = impl_from_WMPStream(iface);
    return This->stream->Read(This->stream, buf, len);
}

static JXR_ERR wmp_stream_Write(struct WMPStream *iface, const void *buf, size_t len)
{
    struct decoder_proxy *This = impl_from_WMPStream(iface);
    return This->stream->Write(This->stream, buf, len);
}

static JXR_ERR wmp_stream_SetPos(struct WMPStream *iface, size_t pos)
{
    struct decoder_proxy *This = impl_from_WMPStream(iface);
    return This->stream->SetPos(This->stream, pos);
}

static JXR_ERR wmp_stream_GetPos(struct WMPStream *iface, size_t *pos)
{
    struct decoder_proxy *This = impl_from_WMPStream(iface);
    return This->stream->GetPos(This->stream, pos);
}

static JXR_ERR CDECL decoder_Initialize(struct jxrlib_decoder *iface, struct jxrlib_stream *stream)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    if (This->stream) return WMP_errFail;
    This->stream = stream;
    return This->decoder->Initialize(This->decoder, &This->WMPStream_iface);
}

static JXR_ERR CDECL decoder_GetPixelFormat(struct jxrlib_decoder *iface, GUID *guid)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    return This->decoder->GetPixelFormat(This->decoder, guid);
}

static JXR_ERR CDECL decoder_GetSize(struct jxrlib_decoder *iface, INT32 *width, INT32 *height)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    return This->decoder->GetSize(This->decoder, width, height);
}

static JXR_ERR CDECL decoder_GetResolution(struct jxrlib_decoder *iface, FLOAT *resx, FLOAT *resy)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    return This->decoder->GetResolution(This->decoder, resx, resy);
}

static JXR_ERR CDECL decoder_Copy(struct jxrlib_decoder *iface, const WICRect *rect, UINT8 *buffer, UINT32 stride)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    PKRect pkrect;
    if (!rect) return WMP_errFail;
    pkrect.X = rect->X;
    pkrect.Y = rect->Y;
    pkrect.Width = rect->Width;
    pkrect.Height = rect->Height;
    return This->decoder->Copy(This->decoder, &pkrect, buffer, stride);
}

static JXR_ERR CDECL decoder_GetFrameCount(struct jxrlib_decoder *iface, UINT32 *count)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    return This->decoder->GetFrameCount(This->decoder, count);
}

static JXR_ERR CDECL decoder_SelectFrame(struct jxrlib_decoder *iface, UINT32 index)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(iface);
    return This->decoder->SelectFrame(This->decoder, index);
}

static JXR_ERR CDECL decoder_Release(struct jxrlib_decoder **iface)
{
    struct decoder_proxy *This = impl_from_jxrlib_decoder(*iface);
    JXR_ERR err;

    if ((err = This->decoder->Release(&This->decoder))) return err;

    free(This);
    *iface = NULL;

    return err;
}

static HRESULT CDECL jxrlib_decoder_create(struct jxrlib_decoder **out)
{
    struct decoder_proxy *This;
    PKImageDecode *decoder;

    if (pPKImageDecode_Create_WMP(&decoder)) return E_FAIL;
    if (!(This = malloc(sizeof(*This)))) return E_OUTOFMEMORY;

    This->decoder_iface.Initialize = &decoder_Initialize;
    This->decoder_iface.GetPixelFormat = &decoder_GetPixelFormat;
    This->decoder_iface.GetSize = &decoder_GetSize;
    This->decoder_iface.GetResolution = &decoder_GetResolution;
    This->decoder_iface.Copy = &decoder_Copy;
    This->decoder_iface.GetFrameCount = &decoder_GetFrameCount;
    This->decoder_iface.SelectFrame = &decoder_SelectFrame;
    This->decoder_iface.Release = &decoder_Release;

    This->WMPStream_iface.Close = &wmp_stream_Close;
    This->WMPStream_iface.EOS = &wmp_stream_EOS;
    This->WMPStream_iface.Read = &wmp_stream_Read;
    This->WMPStream_iface.Write = &wmp_stream_Write;
    This->WMPStream_iface.SetPos = &wmp_stream_SetPos;
    This->WMPStream_iface.GetPos = &wmp_stream_GetPos;

    This->decoder = decoder;
    This->stream = NULL;

    *out = &This->decoder_iface;
    return S_OK;
}
#else
static HRESULT CDECL jxrlib_decoder_create(struct jxrlib_decoder **out)
{
    FIXME("JPEG-XR support not compiled in!\n");
    return E_NOTIMPL;
}
#endif

static const struct jxrlib_funcs jxrlib_funcs =
{
    jxrlib_decoder_create,
};

NTSTATUS CDECL __wine_init_unix_lib(HMODULE module, DWORD reason, const void *ptr_in, void *ptr_out)
{
    TRACE("module %p, reason %d, ptr_in %p, ptr_out %p\n", module, reason, ptr_in, ptr_out);

    if (reason != DLL_PROCESS_ATTACH) return STATUS_SUCCESS;

#ifdef SONAME_LIBJXRGLUE
    if (!(libjxrglue = dlopen(SONAME_LIBJXRGLUE, RTLD_NOW)))
    {
        WARN("failed to load %s\n", SONAME_LIBJXRGLUE);
        return STATUS_DLL_NOT_FOUND;
    }

    if (!(pPKImageDecode_Create_WMP = dlsym(libjxrglue, "PKImageDecode_Create_WMP")))
    {
        ERR("unable to find PKImageDecode_Create_WMP in %s!\n", SONAME_LIBJXRGLUE);
        return FALSE;
    }
#endif

    *(const struct jxrlib_funcs **)ptr_out = &jxrlib_funcs;
    return STATUS_SUCCESS;
}
