/*
 * Copyright 2017 Vincent Povirk for CodeWeavers
 * Copyright 2020 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>

#define COBJMACROS

#include "windef.h"
#include "winbase.h"
#include "winternl.h"
#include "objbase.h"
#include "rpcproxy.h"
#include "wincodecsdk.h"

#include "wine/debug.h"

#include "wmphoto_private.h"

WINE_DEFAULT_DEBUG_CHANNEL(wincodecs);

static const struct jxrlib_funcs *jxrlib_funcs;

static inline const char *debug_wic_rect(const WICRect *rect)
{
    if (!rect) return "(null)";
    return wine_dbg_sprintf("(%u,%u)-(%u,%u)", rect->X, rect->Y, rect->Width, rect->Height);
}

extern HRESULT copy_pixels(UINT bpp, const BYTE *srcbuffer, UINT srcwidth, UINT srcheight, INT srcstride,
                           const WICRect *rc, UINT dststride, UINT dstbuffersize, BYTE *dstbuffer) DECLSPEC_HIDDEN;

static const struct
{
    const WICPixelFormatGUID *format;
    UINT bpp;
} pixel_format_bpp[] =
{
    {&GUID_WICPixelFormat128bppRGBAFixedPoint, 128},
    {&GUID_WICPixelFormat128bppRGBAFloat, 128},
    {&GUID_WICPixelFormat128bppRGBFloat, 128},
    {&GUID_WICPixelFormat16bppBGR555, 16},
    {&GUID_WICPixelFormat16bppBGR565, 16},
    {&GUID_WICPixelFormat16bppGray, 16},
    {&GUID_WICPixelFormat16bppGrayFixedPoint, 16},
    {&GUID_WICPixelFormat16bppGrayHalf, 16},
    {&GUID_WICPixelFormat24bppBGR, 24},
    {&GUID_WICPixelFormat24bppRGB, 24},
    {&GUID_WICPixelFormat32bppBGR, 32},
    {&GUID_WICPixelFormat32bppBGR101010, 32},
    {&GUID_WICPixelFormat32bppBGRA, 32},
    {&GUID_WICPixelFormat32bppCMYK, 32},
    {&GUID_WICPixelFormat32bppGrayFixedPoint, 32},
    {&GUID_WICPixelFormat32bppGrayFloat, 32},
    {&GUID_WICPixelFormat32bppRGBE, 32},
    {&GUID_WICPixelFormat40bppCMYKAlpha, 40},
    {&GUID_WICPixelFormat48bppRGB, 48},
    {&GUID_WICPixelFormat48bppRGBFixedPoint, 48},
    {&GUID_WICPixelFormat48bppRGBHalf, 48},
    {&GUID_WICPixelFormat64bppCMYK, 64},
    {&GUID_WICPixelFormat64bppRGBA, 64},
    {&GUID_WICPixelFormat64bppRGBAFixedPoint, 64},
    {&GUID_WICPixelFormat64bppRGBAHalf, 64},
    {&GUID_WICPixelFormat80bppCMYKAlpha, 80},
    {&GUID_WICPixelFormat8bppGray, 8},
    {&GUID_WICPixelFormat96bppRGBFixedPoint, 96},
    {&GUID_WICPixelFormatBlackWhite, 1},
};

static inline UINT pixel_format_get_bpp(const WICPixelFormatGUID *format)
{
    int i;
    for (i = 0; i < ARRAY_SIZE(pixel_format_bpp); ++i)
        if (IsEqualGUID(format, pixel_format_bpp[i].format)) return pixel_format_bpp[i].bpp;
    return 0;
}

struct stream_proxy
{
    struct jxrlib_stream stream_iface;
    IStream *stream;
};

static inline struct stream_proxy *impl_from_jxrlib_stream(struct jxrlib_stream *iface)
{
    return CONTAINING_RECORD(iface, struct stream_proxy, stream_iface);
}

static JXR_ERR CDECL stream_Read(struct jxrlib_stream *iface, void *buf, size_t len)
{
    struct stream_proxy *This = impl_from_jxrlib_stream(iface);
    ULONG count;

    TRACE("iface %p, buf %p, len %Ix.\n", iface, buf, len);

    if (FAILED(IStream_Read(This->stream, buf, len, &count)) || count < len) return WMP_errFileIO;
    return WMP_errSuccess;
}

static JXR_ERR CDECL stream_Write(struct jxrlib_stream *iface, const void *buf, size_t len)
{
    struct stream_proxy *This = impl_from_jxrlib_stream(iface);
    ULONG count;

    TRACE("iface %p, buf %p, len %Ix.\n", iface, buf, len);

    if (FAILED(IStream_Write(This->stream, buf, len, &count)) || count < len) return WMP_errFileIO;
    return WMP_errSuccess;
}

static JXR_ERR CDECL stream_SetPos(struct jxrlib_stream *iface, size_t pos)
{
    struct stream_proxy *This = impl_from_jxrlib_stream(iface);
    LARGE_INTEGER move;

    TRACE("iface %p, pos %Ix.\n", iface, pos);

    move.QuadPart = pos;
    if (FAILED(IStream_Seek(This->stream, move, STREAM_SEEK_SET, NULL))) return WMP_errFileIO;
    return WMP_errSuccess;
}

static JXR_ERR CDECL stream_GetPos(struct jxrlib_stream *iface, size_t *pos)
{
    struct stream_proxy *This = impl_from_jxrlib_stream(iface);
    ULARGE_INTEGER curr;
    LARGE_INTEGER move;

    TRACE("iface %p, pos %p.\n", iface, pos);

    move.QuadPart = 0;
    if (FAILED(IStream_Seek(This->stream, move, STREAM_SEEK_CUR, &curr))) return WMP_errFileIO;
    *pos = curr.QuadPart;

    return WMP_errSuccess;
}

static HRESULT jxrlib_stream_create(IStream *stream, struct jxrlib_stream **out)
{
    struct stream_proxy *This;
    LARGE_INTEGER seek;
    HRESULT hr;

    seek.QuadPart = 0;
    if (FAILED(hr = IStream_Seek(stream, seek, STREAM_SEEK_SET, NULL))) return hr;
    if (!(This = HeapAlloc(GetProcessHeap(), 0, sizeof(*This)))) return E_OUTOFMEMORY;

    This->stream_iface.Read = &stream_Read;
    This->stream_iface.Write = &stream_Write;
    This->stream_iface.SetPos = &stream_SetPos;
    This->stream_iface.GetPos = &stream_GetPos;
    This->stream = stream;
    IStream_AddRef(stream);

    *out = &This->stream_iface;
    return S_OK;
}

static void jxrlib_stream_release(struct jxrlib_stream *iface)
{
    struct stream_proxy *This = impl_from_jxrlib_stream(iface);
    IStream_Release(This->stream);
    HeapFree(GetProcessHeap(), 0, This);
}

struct wmp_decoder_frame
{
    IWICBitmapFrameDecode IWICBitmapFrameDecode_iface;
    LONG ref;
    WICPixelFormatGUID format;
    UINT bpp;
    UINT stride;
    WICRect rect;
    float resx, resy;
    void *image_data;
};

static inline struct wmp_decoder_frame *impl_from_IWICBitmapFrameDecode(IWICBitmapFrameDecode *iface)
{
    return CONTAINING_RECORD(iface, struct wmp_decoder_frame, IWICBitmapFrameDecode_iface);
}

static HRESULT WINAPI wmp_decoder_frame_QueryInterface(IWICBitmapFrameDecode *iface, REFIID iid, void **out)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);

    TRACE("iface %p, iid %s, out %p.\n", iface, debugstr_guid(iid), out);

    if (!out) return E_INVALIDARG;

    *out = NULL;
    if (!IsEqualIID(&IID_IUnknown, iid) &&
        !IsEqualIID(&IID_IWICBitmapSource, iid) &&
        !IsEqualIID(&IID_IWICBitmapFrameDecode, iid))
        return E_NOINTERFACE;

    *out = &This->IWICBitmapFrameDecode_iface;
    IUnknown_AddRef((IUnknown *)*out);
    return S_OK;
}

static ULONG WINAPI wmp_decoder_frame_AddRef(IWICBitmapFrameDecode *iface)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    ULONG ref = InterlockedIncrement(&This->ref);
    TRACE("iface %p -> ref %u.\n", iface, ref);
    return ref;
}

static ULONG WINAPI wmp_decoder_frame_Release(IWICBitmapFrameDecode *iface)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    ULONG ref = InterlockedDecrement(&This->ref);
    TRACE("iface %p -> ref %u.\n", iface, ref);

    if (ref == 0)
    {
        HeapFree(GetProcessHeap(), 0, This->image_data);
        HeapFree(GetProcessHeap(), 0, This);
    }

    return ref;
}

static HRESULT WINAPI wmp_decoder_frame_GetSize(IWICBitmapFrameDecode *iface, UINT *width, UINT *height)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    TRACE("iface %p, width %p, height %p.\n", iface, width, height);
    *width = This->rect.Width;
    *height = This->rect.Height;
    return S_OK;
}

static HRESULT WINAPI wmp_decoder_frame_GetPixelFormat(IWICBitmapFrameDecode *iface, WICPixelFormatGUID *format)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    TRACE("iface %p, format %p.\n", iface, format);
    *format = This->format;
    return S_OK;
}

static HRESULT WINAPI wmp_decoder_frame_GetResolution(IWICBitmapFrameDecode *iface, double *dpix, double *dpiy)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    TRACE("iface %p, dpix %p, dpiy %p.\n", iface, dpix, dpiy);
    *dpix = This->resx;
    *dpiy = This->resy;
    return S_OK;
}

static HRESULT WINAPI wmp_decoder_frame_CopyPalette(IWICBitmapFrameDecode *iface, IWICPalette *palette)
{
    TRACE("iface %p, palette %p.\n", iface, palette);
    return WINCODEC_ERR_PALETTEUNAVAILABLE;
}

static HRESULT WINAPI wmp_decoder_frame_CopyPixels(IWICBitmapFrameDecode *iface, const WICRect *rect,
                                                   UINT stride, UINT bufsize, BYTE *buffer)
{
    struct wmp_decoder_frame *This = impl_from_IWICBitmapFrameDecode(iface);
    TRACE("iface %p, rect %p, stride %u, bufsize %u, buffer %p.\n", iface, debug_wic_rect(rect),
          stride, bufsize, buffer);
    return copy_pixels(This->bpp, This->image_data, This->rect.Width, This->rect.Height,
                       This->stride, rect, stride, bufsize, buffer);
}

static HRESULT WINAPI wmp_decoder_frame_GetMetadataQueryReader(IWICBitmapFrameDecode *iface,
                                                               IWICMetadataQueryReader **reader)
{
    FIXME("iface %p, reader %p, stub!\n", iface, reader);
    if (!reader) return E_INVALIDARG;
    *reader = NULL;
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_frame_GetColorContexts(IWICBitmapFrameDecode *iface, UINT maxcount,
                                                         IWICColorContext **contexts, UINT *count)
{
    FIXME("iface %p, maxcount %u, contexts %p, count %p, stub\n", iface, maxcount, contexts, count);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_frame_GetThumbnail(IWICBitmapFrameDecode *iface, IWICBitmapSource **thumbnail)
{
    FIXME("iface %p, thumbnail %p, stub!\n", iface, thumbnail);
    return WINCODEC_ERR_CODECNOTHUMBNAIL;
}

static const IWICBitmapFrameDecodeVtbl wmp_decoder_frame_vtbl =
{
    /* IUnknown methods */
    wmp_decoder_frame_QueryInterface,
    wmp_decoder_frame_AddRef,
    wmp_decoder_frame_Release,
    /* IWICBitmapSource methods */
    wmp_decoder_frame_GetSize,
    wmp_decoder_frame_GetPixelFormat,
    wmp_decoder_frame_GetResolution,
    wmp_decoder_frame_CopyPalette,
    wmp_decoder_frame_CopyPixels,
    /* IWICBitmapFrameDecode methods */
    wmp_decoder_frame_GetMetadataQueryReader,
    wmp_decoder_frame_GetColorContexts,
    wmp_decoder_frame_GetThumbnail
};

static HRESULT wmp_decoder_frame_create(IStream *stream, UINT frame, IWICBitmapFrameDecode **out)
{
    struct wmp_decoder_frame *This = NULL;
    struct jxrlib_decoder *jxrlib_decoder = NULL;
    struct jxrlib_stream *jxrlib_stream = NULL;
    HRESULT hr;

    TRACE("stream %p, frame %u, out %p.\n", stream, frame, out);

    *out = NULL;
    if (FAILED(hr = jxrlib_stream_create(stream, &jxrlib_stream))) goto failed;
    if (FAILED(hr = jxrlib_funcs->decoder_create(&jxrlib_decoder))) goto failed;

    hr = E_FAIL;
    if (jxrlib_decoder->Initialize(jxrlib_decoder, jxrlib_stream)) goto failed;
    if (jxrlib_decoder->SelectFrame(jxrlib_decoder, frame)) goto failed;

    hr = E_OUTOFMEMORY;
    if (!(This = HeapAlloc(GetProcessHeap(), 0, sizeof(struct wmp_decoder_frame)))) goto failed;
    This->IWICBitmapFrameDecode_iface.lpVtbl = &wmp_decoder_frame_vtbl;
    This->ref = 1;

    hr = E_FAIL;
    if (jxrlib_decoder->GetPixelFormat(jxrlib_decoder, &This->format)) goto failed;
    if (jxrlib_decoder->GetSize(jxrlib_decoder, &This->rect.Width, &This->rect.Height)) goto failed;
    if (jxrlib_decoder->GetResolution(jxrlib_decoder, &This->resx, &This->resy)) goto failed;

    if (!(This->bpp = pixel_format_get_bpp(&This->format))) goto failed;
    This->stride = (This->rect.Width * This->bpp + 7) / 8;
    if (!(This->image_data = HeapAlloc(GetProcessHeap(), 0, This->rect.Height * This->stride))) goto failed;
    if (jxrlib_decoder->Copy(jxrlib_decoder, &This->rect, This->image_data, This->stride)) goto failed;

    *out = &This->IWICBitmapFrameDecode_iface;
    return S_OK;

failed:
    if (This && This->image_data) HeapFree(GetProcessHeap(), 0, This->image_data);
    if (This) HeapFree(GetProcessHeap(), 0, This);
    if (jxrlib_decoder) jxrlib_decoder->Release(&jxrlib_decoder);
    if (jxrlib_stream) jxrlib_stream_release(jxrlib_stream);
    return hr;
}

struct wmp_decoder
{
    IWICBitmapDecoder IWICBitmapDecoder_iface;
    IWICBitmapFrameDecode IWICBitmapFrameDecode_iface;
    LONG ref;
    IStream *stream;
    UINT frame_count;
    IWICBitmapFrameDecode *frame;
    CRITICAL_SECTION lock;
};

static inline struct wmp_decoder *impl_from_IWICBitmapDecoder(IWICBitmapDecoder *iface)
{
    return CONTAINING_RECORD(iface, struct wmp_decoder, IWICBitmapDecoder_iface);
}

static HRESULT WINAPI wmp_decoder_QueryInterface(IWICBitmapDecoder *iface, REFIID iid, void **out)
{
    struct wmp_decoder *This = impl_from_IWICBitmapDecoder(iface);

    TRACE("iface %p, iid %s, out %p.\n", iface, debugstr_guid(iid), out);

    if (!out) return E_INVALIDARG;

    *out = NULL;
    if (!IsEqualIID(&IID_IUnknown, iid) && !IsEqualIID(&IID_IWICBitmapDecoder, iid))
        return E_NOINTERFACE;

    *out = &This->IWICBitmapDecoder_iface;
    IUnknown_AddRef((IUnknown *)*out);
    return S_OK;
}

static ULONG WINAPI wmp_decoder_AddRef(IWICBitmapDecoder *iface)
{
    struct wmp_decoder *This = impl_from_IWICBitmapDecoder(iface);
    ULONG ref = InterlockedIncrement(&This->ref);
    TRACE("iface %p -> ref %u.\n", iface, ref);
    return ref;
}

static ULONG WINAPI wmp_decoder_Release(IWICBitmapDecoder *iface)
{
    struct wmp_decoder *This = impl_from_IWICBitmapDecoder(iface);
    ULONG ref = InterlockedDecrement(&This->ref);
    TRACE("iface %p -> ref %u.\n", iface, ref);

    if (ref == 0)
    {
        This->lock.DebugInfo->Spare[0] = 0;
        DeleteCriticalSection(&This->lock);
        if (This->frame) IWICBitmapFrameDecode_Release(This->frame);
        if (This->stream) IStream_Release(This->stream);
        HeapFree(GetProcessHeap(), 0, This);
    }

    return ref;
}

static HRESULT WINAPI wmp_decoder_QueryCapability(IWICBitmapDecoder *iface, IStream *stream, DWORD *capability)
{
    FIXME("iface %p, stream %p, capability %p, stub!\n", iface, stream, capability);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_Initialize(IWICBitmapDecoder *iface, IStream *stream, WICDecodeOptions options)
{
    struct wmp_decoder *This = impl_from_IWICBitmapDecoder(iface);
    struct jxrlib_decoder *jxrlib_decoder = NULL;
    struct jxrlib_stream *jxrlib_stream = NULL;
    HRESULT hr;

    TRACE("iface %p, stream %p, options %u.\n", iface, stream, options);

    EnterCriticalSection(&This->lock);

    if (This->stream) hr = WINCODEC_ERR_WRONGSTATE;
    else
    {
        if (FAILED(hr = jxrlib_stream_create(stream, &jxrlib_stream))) goto done;
        if (FAILED(hr = jxrlib_funcs->decoder_create(&jxrlib_decoder))) goto done;

        hr = E_FAIL;
        if (jxrlib_decoder->Initialize(jxrlib_decoder, jxrlib_stream)) goto done;
        if (jxrlib_decoder->GetFrameCount(jxrlib_decoder, &This->frame_count)) goto done;
        if (This->frame_count > 1) FIXME("multi frame JPEG-XR not implemented\n");

        IStream_AddRef(stream);
        This->stream = stream;
        hr = S_OK;
    }

done:
    if (jxrlib_decoder) jxrlib_decoder->Release(&jxrlib_decoder);
    if (jxrlib_stream) jxrlib_stream_release(jxrlib_stream);
    LeaveCriticalSection(&This->lock);
    return hr;
}

static HRESULT WINAPI wmp_decoder_GetContainerFormat(IWICBitmapDecoder *iface, GUID *format)
{
    memcpy(format, &GUID_ContainerFormatWmp, sizeof(GUID));
    return S_OK;
}

static HRESULT WINAPI wmp_decoder_GetDecoderInfo(IWICBitmapDecoder *iface, IWICBitmapDecoderInfo **info)
{
    FIXME("iface %p, info %p, stub!\n", iface, info);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_CopyPalette(IWICBitmapDecoder *iface, IWICPalette *palette)
{
    TRACE("iface %p, palette %p.\n", iface, palette);
    return WINCODEC_ERR_PALETTEUNAVAILABLE;
}

static HRESULT WINAPI wmp_decoder_GetMetadataQueryReader(IWICBitmapDecoder *iface, IWICMetadataQueryReader **reader)
{
    FIXME("iface %p, reader %p, stub!\n", iface, reader);
    if (!reader) return E_INVALIDARG;
    *reader = NULL;
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_GetPreview(IWICBitmapDecoder *iface, IWICBitmapSource **source)
{
    FIXME("iface %p, source %p, stub!\n", iface, source);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_GetColorContexts(IWICBitmapDecoder *iface, UINT maxcount,
                                                   IWICColorContext **contexts, UINT *count)
{
    FIXME("iface %p, maxcount %u, contexts %p, count %p, stub!\n", iface, maxcount, contexts, count);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_GetThumbnail(IWICBitmapDecoder *iface, IWICBitmapSource **thumbnail)
{
    FIXME("iface %p, thumbnail %p, stub!\n", iface, thumbnail);
    return E_NOTIMPL;
}

static HRESULT WINAPI wmp_decoder_GetFrameCount(IWICBitmapDecoder *iface, UINT *count)
{
    TRACE("iface %p, count %p.\n", iface, count);
    if (!count) return E_INVALIDARG;
    *count = 1;
    return S_OK;
}

static HRESULT WINAPI wmp_decoder_GetFrame(IWICBitmapDecoder *iface, UINT index, IWICBitmapFrameDecode **frame)
{
    struct wmp_decoder *This = impl_from_IWICBitmapDecoder(iface);
    HRESULT hr;

    TRACE("iface %p, index %u, frame %p.\n", iface, index, frame);

    if (!frame) return E_INVALIDARG;

    EnterCriticalSection(&This->lock);
    if (!This->stream) hr = WINCODEC_ERR_WRONGSTATE;
    else if (index >= This->frame_count) hr = E_INVALIDARG;
    else if (index >= 1) hr = E_NOTIMPL; /* FIXME: Add support for multiple frames */
    else if (This->frame) hr = S_OK;
    else hr = wmp_decoder_frame_create(This->stream, index, &This->frame);

    if (This->frame) IWICBitmapFrameDecode_AddRef(This->frame);
    *frame = This->frame;
    LeaveCriticalSection(&This->lock);

    return hr;
}

static const IWICBitmapDecoderVtbl wmp_decoder_vtbl =
{
    /* IUnknown methods */
    wmp_decoder_QueryInterface,
    wmp_decoder_AddRef,
    wmp_decoder_Release,
    /* IWICBitmapDecoder methods */
    wmp_decoder_QueryCapability,
    wmp_decoder_Initialize,
    wmp_decoder_GetContainerFormat,
    wmp_decoder_GetDecoderInfo,
    wmp_decoder_CopyPalette,
    wmp_decoder_GetMetadataQueryReader,
    wmp_decoder_GetPreview,
    wmp_decoder_GetColorContexts,
    wmp_decoder_GetThumbnail,
    wmp_decoder_GetFrameCount,
    wmp_decoder_GetFrame
};

static HRESULT CDECL wmp_decoder_create(IUnknown *outer, IUnknown **out)
{
    struct wmp_decoder *This;

    TRACE("outer %p, out %p.\n", outer, out);

    *out = NULL;
    This = HeapAlloc(GetProcessHeap(), 0, sizeof(struct wmp_decoder));
    if (!This) return E_OUTOFMEMORY;

    This->IWICBitmapDecoder_iface.lpVtbl = &wmp_decoder_vtbl;
    This->IWICBitmapFrameDecode_iface.lpVtbl = &wmp_decoder_frame_vtbl;
    This->ref = 1;
    This->stream = NULL;
    This->frame = NULL;
    InitializeCriticalSection(&This->lock);
    This->lock.DebugInfo->Spare[0] = (DWORD_PTR)(__FILE__ ": wmp_decoder.lock");

    *out = (IUnknown *)&This->IWICBitmapDecoder_iface;
    return S_OK;
}

struct class_factory
{
    IClassFactory IClassFactory_iface;
    HRESULT (*create_instance)(IUnknown *outer, IUnknown **out);
};

static inline struct class_factory *impl_from_IClassFactory(IClassFactory *iface)
{
    return CONTAINING_RECORD(iface, struct class_factory, IClassFactory_iface);
}

static HRESULT WINAPI class_factory_QueryInterface(IClassFactory *iface, REFIID iid, void **out)
{
    TRACE("iface %p, iid %s, out %p.\n", iface, debugstr_guid(iid), out);

    if (IsEqualGUID(iid, &IID_IUnknown) || IsEqualGUID(iid, &IID_IClassFactory))
    {
        *out = iface;
        IClassFactory_AddRef(iface);
        return S_OK;
    }

    *out = NULL;
    FIXME("%s not implemented.\n", debugstr_guid(iid));
    return E_NOINTERFACE;
}

static ULONG WINAPI class_factory_AddRef(IClassFactory *iface) { return 2; }

static ULONG WINAPI class_factory_Release(IClassFactory *iface) { return 1; }

static HRESULT WINAPI class_factory_CreateInstance(IClassFactory *iface, IUnknown *outer, REFIID iid, void **out)
{
    struct class_factory *factory = impl_from_IClassFactory(iface);
    IUnknown *unk;
    HRESULT hr;

    TRACE("iface %p, outer %p, iid %s, out %p.\n", iface, outer, debugstr_guid(iid), out);

    if (outer && !IsEqualGUID(iid, &IID_IUnknown)) return E_NOINTERFACE;

    *out = NULL;
    if (SUCCEEDED(hr = factory->create_instance(outer, &unk)))
    {
        hr = IUnknown_QueryInterface(unk, iid, out);
        IUnknown_Release(unk);
    }

    return hr;
}

static HRESULT WINAPI class_factory_LockServer(IClassFactory *iface, BOOL lock)
{
    FIXME("iface %p, lock %d, stub!\n", iface, lock);
    return S_OK;
}

static const IClassFactoryVtbl class_factory_vtbl =
{
    class_factory_QueryInterface,
    class_factory_AddRef,
    class_factory_Release,
    class_factory_CreateInstance,
    class_factory_LockServer,
};

static struct class_factory wmp_decoder_cf = {{&class_factory_vtbl}, wmp_decoder_create};

static HINSTANCE WMPHOTO_hInstance;

BOOL WINAPI DllMain(HINSTANCE instance, DWORD reason, LPVOID reserved)
{
    TRACE("instance %p, reason %d, reserved %p\n", instance, reason, reserved);

    if (FAILED(__wine_init_unix_lib(instance, reason, NULL, &jxrlib_funcs))) return FALSE;

    switch (reason)
    {
    case DLL_PROCESS_ATTACH:
        WMPHOTO_hInstance = instance;
        DisableThreadLibraryCalls(instance);
        break;
    case DLL_WINE_PREATTACH:
        return FALSE; /* prefer native version */
    }

    return TRUE;
}

HRESULT WINAPI DllCanUnloadNow(void)
{
    return S_FALSE;
}

HRESULT WINAPI DllGetClassObject(REFCLSID clsid, REFIID iid, LPVOID *out)
{
    struct class_factory *factory;

    TRACE("clsid %s, iid %s, out %p.\n", debugstr_guid(clsid), debugstr_guid(iid), out);

    if (IsEqualGUID(clsid, &CLSID_WICWmpDecoder)) factory = &wmp_decoder_cf;
    else
    {
        FIXME("%s not implemented.\n", debugstr_guid(clsid));
        return CLASS_E_CLASSNOTAVAILABLE;
    }

    return IClassFactory_QueryInterface(&factory->IClassFactory_iface, iid, out);
}

HRESULT WINAPI DllRegisterServer(void)
{
    return __wine_register_resources( WMPHOTO_hInstance );
}

HRESULT WINAPI DllUnregisterServer(void)
{
    return __wine_unregister_resources( WMPHOTO_hInstance );
}
